@@ -1,20 +1,21 @@
 use crate::{
     resolve_imports, InstantiationError, Resolver, RuntimeError, SerializeError, Tunables,
 };
+use enumset::EnumSet;
 use loupe::MemoryUsage;
 use std::any::Any;
 use std::fs;
 use std::path::Path;
 use std::sync::Arc;
-use wasmer_compiler::Features;
+use wasmer_compiler::{CpuFeature, Features};
 use wasmer_types::entity::{BoxedSlice, PrimaryMap};
 use wasmer_types::{
     DataInitializer, FunctionIndex, InstanceConfig, LocalFunctionIndex, MemoryIndex, ModuleInfo,
     OwnedDataInitializer, SignatureIndex, TableIndex,
 };
 use wasmer_vm::{
-    FunctionBodyPtr, InstanceAllocator, InstanceHandle, MemoryStyle, TableStyle, TrapHandler,
-    VMSharedSignatureIndex, VMTrampoline,
+    FuncDataRegistry, FunctionBodyPtr, InstanceAllocator, InstanceHandle, MemoryStyle, TableStyle,
+    TrapHandler, VMSharedSignatureIndex, VMTrampoline,
 };
 
 /// An `Artifact` is the product that the `Engine`
@@ -43,6 +44,9 @@ pub trait Artifact: Send + Sync + Upcastable + MemoryUsage {
     /// Returns the features for this Artifact
     fn features(&self) -> &Features;
 
+    /// Returns the CPU features for this Artifact
+    fn cpu_features(&self) -> EnumSet<CpuFeature>;
+
     /// Returns the memory styles associated with this `Artifact`.
     fn memory_styles(&self) -> &PrimaryMap<MemoryIndex, MemoryStyle>;
 
@@ -67,6 +71,9 @@ pub trait Artifact: Send + Sync + Upcastable + MemoryUsage {
     /// Returns the associated VM signatures for this `Artifact`.
     fn signatures(&self) -> &BoxedSlice<SignatureIndex, VMSharedSignatureIndex>;
 
+    /// Get the func data registry
+    fn func_data_registry(&self) -> &FuncDataRegistry;
+
     /// Serializes an artifact into bytes
     fn serialize(&self) -> Result<Vec<u8>, SerializeError>;
 
@@ -94,6 +101,16 @@ pub trait Artifact: Send + Sync + Upcastable + MemoryUsage {
         host_state: Box<dyn Any>,
         config: InstanceConfig,
     ) -> Result<InstanceHandle, InstantiationError> {
+        // Validate the CPU features this module was compiled with against the
+        // host CPU features.
+        let host_cpu_features = CpuFeature::for_host();
+        if !host_cpu_features.is_superset(self.cpu_features()) {
+            Err(InstantiationError::CpuFeature(format!(
+                "{:?}",
+                self.cpu_features().difference(host_cpu_features)
+            )))?;
+        }
+
         self.preinstantiate()?;
         let module = self.module();
         let (imports, import_function_envs) = {
